@@ -1,3 +1,4 @@
+#! /bin/jsh
 #
 # Copyright 2013-2014 Yuichiro Moriguchi
 #
@@ -48,7 +49,8 @@ cat << EOF
 #define NINA_HALT_REJECT -72
 #define NINA_BEGIN -2
 #define NINA_DISCARDSTATE 0x4000
-#define NILTOKEN     0
+#define NILTOKEN 0
+#define INDENT 1
 
 #define UNGET_C(o, c)    (nina__unread((struct ${FILENAME}_tag *)(o), (c)))
 #define UNGET(c)         (nina__unread((struct ${FILENAME}_tag *)(__this__), (c)))
@@ -57,6 +59,8 @@ cat << EOF
 #define THROWERROR       longjmp((__this__)->ex, THROW_CODE_ERROR);
 
 #define __stk_accepted_C(o) (((o)->__stk_accepted[(o)->__slen - 1])(o))
+#define __stk_dead_C(o) (((o)->__stk_dead[(o)->__slen - 1])(o))
+#define __stk_emptyTransition_C(o) (((o)->__stk_emptyTransition[(o)->__slen - 1])(o))
 #define __stk_step_C(o, rd, c) (((o)->__stk_step[(o)->__slen - 1])(o, rd, c))
 #define __stk_execaction_C(o, c) (((o)->__stk_execaction[(o)->__slen - 1])(o, c))
 #define __stk_deadState_C(o) (((o)->__stk_deadState[(o)->__slen - 1])(o))
@@ -68,6 +72,8 @@ cat << EOF
 	if((o)->__slen >= NINA_STACKSIZE)  THROWERROR_C(o);\
 	(o)->__stk_step[(o)->__slen]         = ${FILENAME}_ ## tok ## __step;\
 	(o)->__stk_accepted[(o)->__slen]     = ${FILENAME}_ ## tok ## __accepted;\
+	(o)->__stk_emptyTransition[(o)->__slen]     = ${FILENAME}_ ## tok ## __emptyTransition;\
+	(o)->__stk_dead[(o)->__slen]         = ${FILENAME}_ ## tok ## __dead;\
 	(o)->__stk_execaction[(o)->__slen]   = ${FILENAME}_ ## tok ## __execaction;\
 	(o)->__stk_deadState[(o)->__slen]    = ${FILENAME}_ ## tok ## __deadState;\
 	(o)->__stk_finallyState[(o)->__slen] = ${FILENAME}_ ## tok ## __finallyState;\
@@ -76,9 +82,10 @@ cat << EOF
 	(o)->__sts[(o)->__slen++] = (st);\
 }
 
-#define LOOKAHEAD(c)        nina__LOOKAHEAD(__this__, c)
-#define LOOKAHEAD_COMMIT()  nina__LOOKAHEAD_COMMIT(__this__)
-#define LOOKAHEAD_MARK()    nina__LOOKAHEAD_MARK(__this__)
+#define LOOKAHEAD(c)          nina__LOOKAHEAD(__this__, c)
+#define LOOKAHEAD_COMMIT()    nina__LOOKAHEAD_COMMIT(__this__)
+#define LOOKAHEAD_MARK()      nina__LOOKAHEAD_MARK(__this__)
+#define LOOKAHEAD_MARK_INIT() nina__LOOKAHEAD_MARK_INIT(__this__)
 #define DYNAMIC_STARTWITH(a, c)  ${FILENAME}__startWith(a, c)
 #define INIT_DYNAMIC(o, a, st)  ${FILENAME}__initDyanmic(o, a, st)
 #define NINAINITLIST  nina__initlist(__this__)
@@ -97,6 +104,8 @@ do
   echo "static int ${FILENAME}_${ch1}__step(struct ${FILENAME}_tag *__this__,"
   echo "		nina_stream_t __rd, int __c__);"
   echo "static int ${FILENAME}_${ch1}__accepted(struct ${FILENAME}_tag *__this__);"
+  echo "static int ${FILENAME}_${ch1}__dead(struct ${FILENAME}_tag *__this__);"
+  echo "static int ${FILENAME}_${ch1}__emptyTransition(struct ${FILENAME}_tag *__this__);"
   echo "static int ${FILENAME}_${ch1}__execaction(struct ${FILENAME}_tag *__this__, int __c__);"
   echo "static int ${FILENAME}_${ch1}__isend(struct ${FILENAME}_tag *__this__);"
   echo "static int ${FILENAME}_${ch1}__deadState(struct ${FILENAME}_tag *__this__);"
@@ -111,6 +120,8 @@ do
   echo "static int ${FILENAME}__LR_${i}__step(struct ${FILENAME}_tag *__this__,"
   echo "		nina_stream_t __rd, int __c__);"
   echo "static int ${FILENAME}__LR_${i}__accepted(struct ${FILENAME}_tag *__this__);"
+  echo "static int ${FILENAME}__LR_${i}__dead(struct ${FILENAME}_tag *__this__);"
+  echo "static int ${FILENAME}__LR_${i}__emptyTransition(struct ${FILENAME}_tag *__this__);"
   echo "static int ${FILENAME}__LR_${i}__execaction(struct ${FILENAME}_tag *__this__, int __c__);"
   echo "static int ${FILENAME}__LR_${i}__isend(struct ${FILENAME}_tag *__this__);"
   echo "static int ${FILENAME}__LR_${i}__deadState(struct ${FILENAME}_tag *__this__);"
@@ -167,19 +178,74 @@ static int __readsys(struct ${FILENAME}_tag *b, nina_stream_t f) {
 		longjmp(b->ex, THROW_CODE_ERROR);
 	}
 }
+
 EOF
 
 function _putreadf {
-  cat << EOF
+  if [ -n "$ENABLE_INDENT" ]; then
+    cat << EOF
+static int nina_$1(struct ${FILENAME}_tag *b, nina_stream_t f) {
+	int c;
 
+	if(b->__indent_unread >= 0) {
+		c = b->__indent_unread;
+		b->__indent_unread = -1;
+	} else if(!b->__is_line_begin) {
+		c = ${readr}(b, f);
+	} else {
+		b->_indent = 0;
+		while((c = ${readr}(b, f)) == ' ' || c == '\t') {
+			b->_indent++;
+		}
+		b->__indent_unread = c;
+		c = c >= 0 ? INDENT : c;
+	}
+
+	b->__is_line_begin = c == '\n';
+	if(b->__is_line_begin) {
+		b->_indentBefore = b->_indent;
+	}
+	return c;
+}
+EOF
+  else
+    cat << EOF
 static int nina_$1(struct ${FILENAME}_tag *b, nina_stream_t f) {
 	return ${readr}(b, f);
 }
 EOF
+  fi
 }
 
 if [ -z "$LEXER" ]; then
-  _putreadf _read1
+  _putreadf _read1l
+  cat << EOF
+
+static int nina__read1(struct ${FILENAME}_tag *b, nina_stream_t stream) {
+	int c;
+
+	if(b->_unreadl != -1) {
+		c = b->_unreadl;
+		b->_unreadl = -1;
+EOF
+  if [ -z "$TYPE" -a -z "$NEWLINEMODE" ]; then
+    cat << EOF
+	} else if((c = nina__read1l(b, stream)) == '\r' && (c = nina__read1l(b, stream)) != '\n') {
+		b->_unreadl = c;
+		c = '\r';
+	}
+EOF
+  else
+    cat << EOF
+	} else {
+		c = nina__read1l(b, stream);
+	}
+EOF
+  fi
+  cat << EOF
+	return c;
+}
+EOF
 else
   _putreadf _read1l
   nina_template.nfa.c.sub2.sh "$LEXER" "$LEXER" "int " "$FILENAME" "0" "$LCTYPE"
@@ -188,12 +254,25 @@ else
 static int nina__read1ul(struct ${FILENAME}_tag *b, nina_stream_t stream) {
 	int c;
 
-	if(b->_unreadl == -1) {
-		c = nina__read1l(b, stream);
-	} else {
+	if(b->_unreadl != -1) {
 		c = b->_unreadl;
 		b->_unreadl = -1;
+EOF
+  if [ -z "$LTYPE" -a -z "$NEWLINEMODE" ]; then
+    cat << EOF
+	} else if((c = nina__read1l(b, stream)) == '\r' && (c = nina__read1l(b, stream)) != '\n') {
+		b->_unreadl = c;
+		c = '\r';
 	}
+EOF
+  else
+    cat << EOF
+	} else {
+		c = nina__read1l(b, stream);
+	}
+EOF
+  fi
+  cat << EOF
 	return c;
 }
 
@@ -331,6 +410,8 @@ static void nina__GETCC(struct ${FILENAME}_tag *c, ${FILENAME}_continuation *b)
 	memcpy(b->__sts, c->__sts, sizeof(b->__sts));
 	memcpy(b->__stk_step, c->__stk_step, sizeof(b->__stk_step));
 	memcpy(b->__stk_accepted, c->__stk_accepted, sizeof(b->__stk_accepted));
+	memcpy(b->__stk_dead, c->__stk_dead, sizeof(b->__stk_dead));
+	memcpy(b->__stk_emptyTransition, c->__stk_emptyTransition, sizeof(b->__stk_emptyTransition));
 	memcpy(b->__stk_execaction, c->__stk_execaction, sizeof(b->__stk_execaction));
 	memcpy(b->__stk_deadState, c->__stk_deadState, sizeof(b->__stk_deadState));
 	memcpy(b->__stk_finallyState, c->__stk_finallyState, sizeof(b->__stk_finallyState));
@@ -344,6 +425,8 @@ static void nina__SETCC(struct ${FILENAME}_tag *b, ${FILENAME}_continuation *c)
 	memcpy(b->__sts, c->__sts, sizeof(b->__sts));
 	memcpy(b->__stk_step, c->__stk_step, sizeof(b->__stk_step));
 	memcpy(b->__stk_accepted, c->__stk_accepted, sizeof(b->__stk_accepted));
+	memcpy(b->__stk_dead, c->__stk_dead, sizeof(b->__stk_dead));
+	memcpy(b->__stk_emptyTransition, c->__stk_emptyTransition, sizeof(b->__stk_emptyTransition));
 	memcpy(b->__stk_execaction, c->__stk_execaction, sizeof(b->__stk_execaction));
 	memcpy(b->__stk_deadState, c->__stk_deadState, sizeof(b->__stk_deadState));
 	memcpy(b->__stk_finallyState, c->__stk_finallyState, sizeof(b->__stk_finallyState));
@@ -404,6 +487,12 @@ static void nina__LOOKAHEAD(struct ${FILENAME}_tag *b, $CTYPE c) {
 }
 
 static void nina__copy_lookahead(struct ${FILENAME}_tag *b, int p) {
+	if(b->__lookahead_ptr > 0) {
+		memcpy(b->__lookahead, b->__lookahead + b->__lookahead_ptr,
+				b->__lookahead_length - b->__lookaheadw_ptr);
+		b->__lookahead_length -= b->__lookahead_ptr;
+	}
+
 	if(b->__lookahead == NULL) {
 		b->__lookahead = b->__lookahead_buf;
 		b->__lookahead_length = b->__lookaheadw_ptr;
@@ -442,6 +531,10 @@ static void nina__LOOKAHEAD_MARK(struct ${FILENAME}_tag *b) {
 	b->__lookahead_mark = b->__lookaheadw_ptr;
 }
 
+static void nina__LOOKAHEAD_MARK_INIT(struct ${FILENAME}_tag *b) {
+	b->__lookahead_mark = 0;
+}
+
 static int nina__read(struct ${FILENAME}_tag *b, nina_stream_t f) {
 	int c;
 
@@ -603,6 +696,14 @@ static int ${FILENAME}_dynamic__accepted(struct ${FILENAME}_tag *__this__) {
 	return NINA_TRUE;
 }
 
+static int ${FILENAME}_dynamic__dead(struct ${FILENAME}_tag *__this__) {
+	return NINA_FALSE;
+}
+
+static int ${FILENAME}_dynamic__emptyTransition(struct ${FILENAME}_tag *__this__) {
+	return NINA_FALSE;
+}
+
 static int ${FILENAME}_dynamic__execaction(struct ${FILENAME}_tag *__this__, $CTYPE c) {
 	return 1;
 }
@@ -681,6 +782,16 @@ do
   print_recover $i
   echo '}'
   echo
+
+  echo "static int ${FILENAME}_${ch1}__dead(struct ${FILENAME}_tag *__this__) {"
+  print_deads $i
+  echo '}'
+  echo
+
+  echo "static int ${FILENAME}_${ch1}__emptyTransition(struct ${FILENAME}_tag *__this__) {"
+  print_empty_transitions $i
+  echo '}'
+  echo
 done
 
 for i in $LRPARSERS
@@ -796,7 +907,7 @@ static int ${FILENAME}_parse_int(struct ${FILENAME}_tag *o,
 
 	do {
 		if(c < NINA_EOF) {
-			// do nothing
+			c = 1;
 		} else if(o->__trapped_signal > 0) {
 			__logprint("trapped a signal: ", o->__trapped_signal);
 			return NINA_SIGNAL;
@@ -851,7 +962,29 @@ static int ${FILENAME}_parse_int(struct ${FILENAME}_tag *o,
 			o->failchar = c;
 			return NINA_FAIL;
 		}
-		c = nina__read(o, rd);
+
+		if(__stk_emptyTransition_C(o)) {
+			/* do nothing */
+		} else if(!__stk_dead_C(o)) {
+			c = nina__read(o, rd);
+		} else if(b) {
+			__logprints("accept ", __stk_string_C(o));
+			return NINA_ACCEPT;
+		} else if(o->__lookaheadw_ptr >= 0) {
+			__logprint("match failed: try lookahead: ", c);
+			nina__LOOKAHEAD_RB(o);
+			b = __stk_accepted_C(o);
+		} else if(nina__BACKTRACK(o, c)) {
+			__logprint("match failed: try backtracking: ", c);
+		} else if(c < 0) {
+			if(!b)  THROWERROR_C(o);
+			return NINA_ACCEPT;
+		} else {
+			__logprint("match failed: ", c);
+			__puttrace();
+			o->failchar = c;
+			return NINA_FAIL;
+		}
 	} while(1);
 }
 
@@ -901,8 +1034,8 @@ static int ${FILENAME}_parse_generic(struct ${FILENAME}_tag *b, nina_stream_t f)
 		} else if(c == NINA_SIGNAL) {
 			s = __stk_string_C(b);
 			while((b->state = __stk_signal_C(b)) < 0) {
+				EXECFINALLY;
 				if(b->__slen-- <= 1) {
-					EXECFINALLY;
 					fprintf(stderr, "${FILENAME}: state %s: signal trapped: %d\n",
 							s, b->__trapped_signal);
 					b->__trapped_signal = 0;
@@ -913,8 +1046,8 @@ static int ${FILENAME}_parse_generic(struct ${FILENAME}_tag *b, nina_stream_t f)
 		} else if(c == NINA_FAIL) {
 			s = __stk_string_C(b);
 			while((b->state = __stk_deadState_C(b)) < 0) {
+				EXECFINALLY;
 				if(b->__slen-- <= 1) {
-					EXECFINALLY;
 					fprintf(stderr,
 							"${FILENAME}:error: state %s: character %c(0x%x)\n",
 							s, b->failchar, b->failchar);
@@ -1032,7 +1165,11 @@ void ${FILENAME}_init(struct ${FILENAME}_tag *b) {
 	b->__lexptr = b->__lexbuf;
 	b->__lexok = NINA_FALSE;
 EOF
-[ -n "$LEXER" ] && echo "	b->_unreadl = -1;"
+[ -n "$ENABLE_INDENT" ] && cat << EOF
+	b->__is_line_begin = NINA_TRUE;
+	b->__indent_unread = -1;
+EOF
+echo "	b->_unreadl = -1;"
 echo '}'
 
 echo
@@ -1041,6 +1178,23 @@ echo "	__nina_signal_automaton = b;"
 print_inittrap
 echo '}'
 
+[ -n "$PUTMAIN" ] && cat << EOF
+int main(int argc, char **args) {
+	${FILENAME} k;
+
+	${FILENAME}_init(&k);
+	k.parse(&k, stdin);
+}
+EOF
+[ -n "$PUTPROMPT" ] && cat << EOF
+int main(int argc, char **args) {
+	${FILENAME} k;
+
+	${FILENAME}_init(&k);
+	k.parse(&k, stdin);
+}
+EOF
+
 #
 # print header file
 #
@@ -1107,6 +1261,8 @@ typedef struct ${FILENAME}_continuation_tag {
 	int __sts[NINA_STACKSIZE];
 	int ((*__stk_step[NINA_STACKSIZE])(struct ${FILENAME}_tag *, nina_stream_t, int));
 	int ((*__stk_accepted[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
+	int ((*__stk_dead[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
+	int ((*__stk_emptyTransition[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
 	int ((*__stk_execaction[NINA_STACKSIZE])(struct ${FILENAME}_tag *, int));
 	int ((*__stk_deadState[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
 	int ((*__stk_finallyState[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
@@ -1148,6 +1304,8 @@ struct ${FILENAME}_tag {
 	int __sts[NINA_STACKSIZE];
 	int ((*__stk_step[NINA_STACKSIZE])(struct ${FILENAME}_tag *, nina_stream_t, int));
 	int ((*__stk_accepted[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
+	int ((*__stk_dead[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
+	int ((*__stk_emptyTransition[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
 	int ((*__stk_execaction[NINA_STACKSIZE])(struct ${FILENAME}_tag *, int));
 	int ((*__stk_deadState[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
 	int ((*__stk_finallyState[NINA_STACKSIZE])(struct ${FILENAME}_tag *));
@@ -1198,8 +1356,18 @@ struct ${FILENAME}_tag {
 
 	$VALTYPE _, *_valp;
 	$VALTYPE _l[VALLISTSIZE];
+
 EOF
-[ -n "$LEXER" ] && echo "	int _unreadl;" >> $hfile
+
+[ -n "$ENABLE_INDENT" ] && cat >> $hfile << EOF
+	int _indent;
+	int _indentBefore;
+	int __is_line_begin;
+	int __indent_unread;
+
+EOF
+
+echo "	int _unreadl;" >> $hfile
 echo '	struct {' >> $hfile
 if [ -s field ]
 then
